#!/usr/bin/env python
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>>Custom Font Tools>>Custom Font Faux 3D v. 0.02
# Version 0.01 Options for optionally adding a decorative layer and creating a 3D style effect using multiple layers with a custom font file
#         Depth, direction of 3D and shading can be modified through the filter dialog 
#         Option for modifing existing kerning
# Version 0.02 New style added with a slightly more perspective view of the 3d layering using scaling
#         Modes to darken 3D layers, blurring and shading values of the 3D layers now available in the dialog
# Version 0.03 Adds preview options to display a character with the current values in a new image


from gimpfu import *
#from array import array
#import math
import random


# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)


def do_stuff(image, layer, style, scale_inc, depth, blur, use_decor, initial_shade, shade_range, darken_mode, direction, kerning_adjustment, do_preview, preview_char):
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
	
    test_for_custom_layer = False
		
    max_width, im_height = proc_do_3D(image, style, scale_inc, depth, blur, use_decor, initial_shade, shade_range, darken_mode, direction, kerning_adjustment, do_preview, preview_char)
	
    if not do_preview: pdb.gimp_image_resize(image, max_width, im_height, 0, 0) # resize canvas to fit characters
	
    if (kerning_adjustment<> 0) and (do_preview == False): # adjust kerning paths if required by dialog
			
    # get lists of vector ids in image
        num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
        loop = 0
        while loop<num_vectors:
		
            # look for recognised kerning path
            kerning_path_name = pdb.gimp_item_get_name(vector_list[loop])
            kerning_path_name = kerning_path_name.decode('utf-8')
			
            if (len(kerning_path_name) == 5 and ".png" in kerning_path_name) or (len(kerning_path_name) == 6 and ".png" in kerning_path_name):
                    
                kerning_vector = vector_list[loop]
                num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(kerning_vector)
                type, num_points, controlpoints, closed  = pdb.gimp_vectors_stroke_get_points(kerning_vector, stroke_ids[0])
                kerning_target_width = controlpoints[6] + depth
                v_position = pdb.gimp_image_get_vectors_position(image, kerning_vector)
                pdb.gimp_image_remove_vectors(image, kerning_vector)
                new_vector_id = pdb.gimp_vectors_new(image, kerning_path_name)
                pdb.gimp_image_insert_vectors(image, new_vector_id, None, v_position)
                points1 = [controlpoints[0],controlpoints[1]] * 3
                points2 = [kerning_target_width,controlpoints[7]] * 3
                points = points1 + points2
                new_vector_id = pdb.gimp_vectors_stroke_new_from_points(new_vector_id ,0, len(points), points, False)
            loop += 1
			
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
	

######################################################################################################################################
	
def proc_do_3D(image, style, scale_inc, depth, blur, use_decor, initial_shade, shade_range, darken_mode, direction, kerning_adjustment, do_preview, preview_char):


    test_for_custom_layer = False
		
    active_layer_id = pdb.gimp_image_get_active_layer(image)
    multi_layer_list = "No non-character layers."
	
    # set mode for 3D layers (increasingly darker)
    if darken_mode == 0:
        darken_mode = 35 # darken only
    elif darken_mode == 1:
        darken_mode = 54 # Luma/luminance darken
    elif darken_mode == 2:
        darken_mode = 30 # Multiply
    elif darken_mode == 3:
        darken_mode = 43 # Burn
    elif darken_mode == 4:
        darken_mode = 53 # Linear burn
	
	
    loop_ptr = 1
    first_char_layer_name = "none"
    num_layers, layer_ids = pdb.gimp_image_get_layers(image)
    loop = num_layers - loop_ptr
    layer_names_list = image.layers
    original_active_layer = image.active_layer
	
	
		
    num_multi_layers,other_layers,num_char_layers = 0,0,0
    visi_loop, first_char_layer  = 0, -1
    im_width = image.width
    im_height = image.height
	
	
    visi_loop = 0

    # loop through layers and turn off visibility of character layers
    while (visi_loop < num_layers and other_layers > 0):
        visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
        visi_name = visi_name.decode('utf-8')
        if visi_name == "space.png" or visi_name == " .png": 
            has_a_space = True
            space_loop = visi_loop
        if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
            # turn off visibility of letter layers
            pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)  
        visi_loop += 1
		
		
    preview_decorate_visibility = False

    visi_loop = 0
    while visi_loop < num_layers:
        # get info about layer we are working on
        layer_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
        char_layer_name = layer_name.decode('utf-8')
        # check to see if the current layer is a custom font character layer
        test_for_custom_layer = fn_test_for_custom_layer(image, layer_names_list[visi_loop], char_layer_name)
			
        if test_for_custom_layer:
            num_char_layers += 1
            if first_char_layer < 0: 
                first_char_layer = visi_loop
                first_char_layer_name = char_layer_name
                layer_names_list[visi_loop].set_offsets(0, 0)
				
        else:
            # look for the presence of decor layers - we use only the top one
            pdb.gimp_image_select_item(image, 2, layer_names_list[visi_loop])
            is_empty = pdb.gimp_selection_is_empty(image)
            if is_empty == False:
                if other_layers == 0:
                        
                    multi_layer_list = [layer_name]
                    preview_decorate_layer = layer_names_list[visi_loop]
                    preview_decorate_visibility = pdb.gimp_item_get_visible(layer_names_list[visi_loop])
                else:
                        
                    multi_layer_list.append(layer_name)
                other_layers += 1
				
            
        if not do_preview: pdb.gimp_item_set_visible(layer_names_list[visi_loop],False) # turn off visibility
        visi_loop += 1
		
    if str(multi_layer_list) == "No non-character layers.": 
        num_multi_layers = 0
        use_decor = False
    else:
        num_multi_layers = len(multi_layer_list)

    # preview character checks
    if preview_char == "": preview_char = "A"
    preview_char = preview_char + ".png"
	
    # check that the preview character is present or use another
    if preview_char in str(layer_names_list):
        prev_pos = str(layer_names_list).find(preview_char)
        prev_loop = str(layer_names_list).count(",",0,prev_pos)
    else:
        loop = 0
        preview_char = pdb.gimp_item_get_name(layer_names_list[loop])
        while preview_char=="space.png" or preview_char in multi_layer_list:
            loop += 1
            preview_char = pdb.gimp_item_get_name(layer_names_list[loop])

    preview_char = preview_char.decode('utf-8')	######## Added to catch higher value unicodes

			
    if not use_decor: num_multi_layers = 1 # ie we are going to copy the existing decorated characters
			
    if do_preview:
        # copy the preview character and paste it as a  new image
		
        preview_layer = pdb.gimp_image_get_layer_by_name(image, preview_char)
        success = pdb.gimp_edit_copy(preview_layer)
        preview_image = pdb.gimp_edit_paste_as_new_image(preview_layer)
        
        ##############################################################
        # begin undo for preview image
        pdb.gimp_image_undo_group_start(preview_image) ##############
		
        preview_display = pdb.gimp_display_new(preview_image)
        preview_num_layers, preview_layer_ids = pdb.gimp_image_get_layers(preview_image)
        preview_layer_names_list = preview_image.layers
		
		
	
    if num_multi_layers>0:
        visi_loop, max_width = 0, 0
        if use_decor: 
            decor_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[0])
            if do_preview:
                # copy decor layer from image and paste to preview image
                decor_copy = pdb.gimp_layer_new(preview_image, decor_layer.width, decor_layer.height, 1, "decor copy", 100, 0, )
                preview_image.add_layer(decor_copy, -1)
                success = pdb.gimp_edit_copy(decor_layer)
                # paste the selection into the floating layer
                floating_sel = pdb.gimp_edit_paste(decor_copy,True)
                # and anchor it
                pdb.gimp_floating_sel_anchor(floating_sel)
                # Re-position the copied layer
                decor_copy.set_offsets(0, 0)
				
        if do_preview:
            preview_loop = 0
            preview_im_width, preview_im_height = preview_image.width, preview_image.height
            while preview_loop < preview_num_layers:
                # get info about layer we are working on for custom font mode
                preview_layer_id = preview_layer_ids[preview_loop]
                preview_layer_name = preview_layer_names_list[preview_loop]

                # cater for unicode characters above normal range
                display_layer_name =  pdb.gimp_item_get_name(preview_layer_names_list[preview_loop])
                display_layer_name = display_layer_name.decode('utf-8')	
	
                # check we have the layer id of the currently active layer
                pdb.gimp_image_set_active_layer(preview_image, preview_layer_name)
                layer_id = pdb.gimp_image_get_active_layer(preview_image)

                # check to see if the current layer is a custom font character layer
                test_for_custom_layer = fn_test_for_custom_layer(preview_image, layer_id, display_layer_name)
	
                if test_for_custom_layer:
                    # ensure we have the active layer and make it visible
                    active_layer_id = pdb.gimp_image_get_active_layer(preview_image)
                    pdb.gimp_item_set_visible(active_layer_id, True)
			
                    # Take 3D direction from dialog
				
                    x_adj, y_adj = 1,1 # SE defaults
                    if direction == 0: x_adj, y_adj = 0,-1 # N
                    if direction == 1: x_adj, y_adj = 1,-1 # NE
                    if direction == 2: x_adj, y_adj = 1,0 # E
                    if direction == 3: x_adj, y_adj = 1,1 # SE
                    if direction == 4: x_adj, y_adj = 0,1 # S
                    if direction == 5: x_adj, y_adj = -1,1 # SW
                    if direction == 6: x_adj, y_adj = -1,0 # W
                    if direction == 7: x_adj, y_adj = -1,-1 # NW
				
                    # Step 1: Apply decor to initial character face if required by dialog
				
                    # make sure that existing layers are positioned at the origin of the image
                    preview_layer_names_list[preview_loop].set_offsets(0, 0)
					
                    # prepare to move layer to random position within the confines of the image size
                    layer_width = preview_layer_names_list[preview_loop].width
                    if layer_width > preview_im_width: layer_width = preview_im_width
                    layer_height = preview_layer_names_list[preview_loop].height
                    if layer_height > preview_im_height: layer_height = preview_im_height
					
                    # make a duplicate of the current character layer and insert
                    layer_position = pdb.gimp_image_get_item_position(preview_image, preview_layer_names_list[preview_loop])
                    layer_copy = pdb.gimp_layer_copy(preview_layer_names_list[preview_loop], True)
                    pdb.gimp_image_insert_layer(preview_image, layer_copy, None, layer_position)
					
					
				
                    if use_decor: preview_layer_names_list, layer_copy = proc_decor(preview_image, preview_im_width, layer_width, preview_im_height, layer_height, layer_copy, decor_layer, preview_layer_names_list, preview_loop)
					
                    # Step 2: Apply 3D effect to current character
					
					
                    # fill shade values taken from dialog
                    my_grey, shade_step = initial_shade, float(shade_range/depth)
                
                    grey_shade = gimpcolor.RGB(my_grey, my_grey, my_grey) # colour value
                    shade_loop_max, shade_loop  = depth, 0
                    decor_char = preview_layer_names_list[preview_loop] # get id of decorated character
                    scale_factor = scale_inc
				
                
                    if blur>0: 
                        # select the contents of the layer
                        pdb.gimp_image_select_item(preview_image, 2, layer_copy)
                        # shrink the selection
                        pdb.gimp_selection_shrink(preview_image, 2)
                        # invert selection
                        pdb.gimp_selection_invert(preview_image)
                        pdb.plug_in_gauss(preview_image, layer_copy, blur, blur, 0)
                        # cancel alpha selection
                        pdb.gimp_selection_none(preview_image)
						
						
                    if style == 0:
                        decor_char = proc_do_3D_translate(preview_image, decor_char, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_factor, blur, shade_loop, my_grey)
				
						
                    elif style == 1:
			
                        decor_char, layer_copy = proc_do_3D_scale(preview_image, decor_char, layer_copy, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_inc, blur, shade_loop, my_grey)
				
                        #pdb.gimp_message(err_list)
					
						
                    pdb.gimp_item_set_visible(decor_char, True)
				
                    if style ==  1 and (not use_decor): pdb.gimp_image_remove_layer(preview_image, layer_copy)
					
					
                    if use_decor: 					
                        pdb.gimp_image_remove_layer(preview_image, decor_copy)
						
                    preview_layer_names_list[preview_loop] = pdb.gimp_image_merge_visible_layers(preview_image, 0)
                    pdb.gimp_item_set_name(preview_layer_names_list[preview_loop], display_layer_name) # rename with character name
				
                    # note new width of character
                    new_width = preview_layer_names_list[preview_loop].width
                    if new_width>max_width: max_width = new_width
                    pdb.gimp_item_set_name(preview_layer_names_list[preview_loop], display_layer_name) # rename with character name
                    pdb.gimp_layer_resize(preview_layer_names_list[preview_loop],new_width,layer_height,0,0) # resize layer to new width
                    # Re-position the layer
                    preview_layer_names_list[preview_loop].set_offsets(0, 0)
                    pdb.gimp_item_set_visible(preview_layer_names_list[preview_loop], False)
			
			
                preview_loop +=1
        else:
            while visi_loop < num_layers:
                # get info about layer we are working on for custom font mode
                layer_id = layer_ids[visi_loop]
                layer_name = layer_names_list[visi_loop]

                # cater for unicode characters above normal range
                display_layer_name =  pdb.gimp_item_get_name(layer_names_list[visi_loop])
                display_layer_name = display_layer_name.decode('utf-8')	
	
                # check we have the layer id of the currently active layer
                pdb.gimp_image_set_active_layer(image, layer_name)
                layer_id = pdb.gimp_image_get_active_layer(image)

                # check to see if the current layer is a custom font character layer
                test_for_custom_layer = fn_test_for_custom_layer(image, layer_id, display_layer_name)
	
	
                if test_for_custom_layer:
                    # ensure we have the active layer and make it visible
                    active_layer_id = pdb.gimp_image_get_active_layer(image)
                    pdb.gimp_item_set_visible(active_layer_id, True)
			
                    # Take 3D direction from dialog
				
                    x_adj, y_adj = 1,1 # SE defaults
                    if direction == 0: x_adj, y_adj = 0,-1 # N
                    if direction == 1: x_adj, y_adj = 1,-1 # NE
                    if direction == 2: x_adj, y_adj = 1,0 # E
                    if direction == 3: x_adj, y_adj = 1,1 # SE
                    if direction == 4: x_adj, y_adj = 0,1 # S
                    if direction == 5: x_adj, y_adj = -1,1 # SW
                    if direction == 6: x_adj, y_adj = -1,0 # W
                    if direction == 7: x_adj, y_adj = -1,-1 # NW
				
                    # Step 1: Apply decor to initial character face if required by dialog
				
                    # make sure that existing layers are positioned at the origin of the image
                    layer_names_list[visi_loop].set_offsets(0, 0)
				
                    # prepare to move layer to random position within the confines of the image size
                    layer_width = layer_names_list[visi_loop].width
                    if layer_width > im_width: layer_width = im_width
                    layer_height = layer_names_list[visi_loop].height
                    if layer_height > im_height: layer_height = im_height
				
					
                    # make a duplicate of the current character layer and insert
                    layer_position = pdb.gimp_image_get_item_position(image, layer_names_list[visi_loop])
                    layer_copy = pdb.gimp_layer_copy(layer_names_list[visi_loop], True)
                    pdb.gimp_image_insert_layer(image, layer_copy, None, layer_position)
				
                    if use_decor: layer_names_list, layer_copy = proc_decor(image, im_width, layer_width, im_height, layer_height, layer_copy, decor_layer, layer_names_list, visi_loop)
					
                    # Step 2: Apply 3D effect to current character
                
                    # fill shade values taken from dialog
                    my_grey, shade_step = initial_shade, float(shade_range/depth)
                
                    grey_shade = gimpcolor.RGB(my_grey, my_grey, my_grey) # colour value
                    shade_loop_max, shade_loop  = depth, 0
                    decor_char = layer_names_list[visi_loop] # get id of decorated character
                    scale_factor = scale_inc
				
                
                    if blur>0: 
                        # select the contents of the layer
                        pdb.gimp_image_select_item(image, 2, layer_copy)
                        # shrink the selection
                        pdb.gimp_selection_shrink(image, 2)
                        # invert selection
                        pdb.gimp_selection_invert(image)
                        pdb.plug_in_gauss(image, layer_copy, blur, blur, 0)
                        # cancel alpha selection
                        pdb.gimp_selection_none(image)
                
				
                    if style == 0:
                        decor_char = proc_do_3D_translate(image, decor_char, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_factor, blur, shade_loop, my_grey)
				
						
                    elif style == 1:
			
                        decor_char, layer_copy = proc_do_3D_scale(image, decor_char, layer_copy, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_inc, blur, shade_loop, my_grey)
				
					
						
                    pdb.gimp_item_set_visible(decor_char, True)
				
                    if style ==  1 and (not use_decor): pdb.gimp_image_remove_layer(image, layer_copy)
                
                    if use_decor: pdb.gimp_item_set_visible(decor_layer, False) # Turn off decor layer visibility
                    layer_names_list[visi_loop] = pdb.gimp_image_merge_visible_layers(image, 0)
                    pdb.gimp_item_set_name(layer_names_list[visi_loop], display_layer_name) # rename with character name
				
                    # note new width of character
                    new_width = layer_names_list[visi_loop].width
                    if new_width>max_width: max_width = new_width
                    pdb.gimp_item_set_name(layer_names_list[visi_loop], display_layer_name) # rename with character name
                    pdb.gimp_layer_resize(layer_names_list[visi_loop],new_width,layer_height,0,0) # resize layer to new width
                    # Re-position the layer
                    layer_names_list[visi_loop].set_offsets(0, 0)
                    pdb.gimp_item_set_visible(layer_names_list[visi_loop], False)
			
                visi_loop += 1
							
				
			
        if use_decor: 
            if not do_preview: pdb.gimp_item_set_visible(decor_layer, False) # Turn off decor layer visibility
			
        if do_preview: 
            pdb.gimp_image_resize(preview_image, max_width, im_height, 0, 0) # resize canvas to fit characters
            preview_layer = pdb.gimp_image_get_layer_by_name(preview_image, preview_char)
            pdb.gimp_item_set_visible(preview_layer, True)
            if use_decor: pdb.gimp_item_set_visible(preview_decorate_layer,  preview_decorate_visibility)
			
            ###########################################
            # Set end undo point for preview image
            pdb.gimp_image_undo_group_end(preview_image)
            ###########################################
		
        if first_char_layer_name <> "none": # the character layer to make visible on completion
            first_char_layer_id = pdb.gimp_image_get_layer_by_name(image, first_char_layer_name)
            pdb.gimp_item_set_visible(first_char_layer_id, True)
            pdb.gimp_image_set_active_layer(image,first_char_layer_id)
				
			
    return max_width, im_height
			
######################################################################################################################################

def proc_decor(image, im_width, layer_width, im_height, layer_height, layer_copy, decor_layer, layer_names_list, visi_loop):

    # set random position for copying
    random.seed()
    random_x = random.randrange(0, (im_width + 1) - layer_width)
    random.seed()
    random_y = random.randrange(0, (im_height + 1) - layer_height)

    # set random position to copy visible layers 
    layer_copy.set_offsets(random_x, random_y)
			
			
    # select the contents of the layer and clear contents
    pdb.gimp_image_select_item(image, 2, layer_copy)
    pdb.gimp_drawable_edit_clear(layer_copy)
			
    # copy the contents of the visible layers into the selection
    pdb.gimp_item_set_visible(decor_layer, True)
    pdb.gimp_selection_grow(image, 1)
    success = pdb.gimp_edit_copy_visible(image)
				
    # turn on layer visibility
    pdb.gimp_item_set_visible(layer_copy, True) 
					
    # paste the selection into the floating layer
    floating_sel = pdb.gimp_edit_paste(layer_copy,True)
    # and anchor it
    pdb.gimp_floating_sel_anchor(floating_sel)
			
    # Re-position the copied layer
    layer_copy.set_offsets(0, 0)
				
    # Make the original character layer visible before we merge down					
    pdb.gimp_item_set_visible( layer_names_list[visi_loop], True)
				
    # cancel alpha selection
    pdb.gimp_selection_none(image)
    # clear the character layer
    pdb.gimp_drawable_edit_clear(layer_names_list[visi_loop])
				
                   
    # Merge down the copied layer to the character layer
    layer_copy = pdb.gimp_image_merge_down(image, layer_copy, 0)
    # Re-set the id of the layer
    layer_names_list[visi_loop] = pdb.gimp_image_get_active_layer(image)
				
    # Turn off layer visibility
    pdb.gimp_item_set_visible(layer_names_list[visi_loop], False) 
	
    return layer_names_list, layer_copy
	
	

######################################################################################################################################


def proc_do_3D_translate(image, decor_char, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_factor, blur, shade_loop, my_grey):

    while shade_loop<shade_loop_max:
        new_grey = (1.0/255) * my_grey
        grey_shade = gimpcolor.RGB(new_grey, new_grey, new_grey)
        pdb.gimp_context_set_foreground(grey_shade)
                    
        if shade_loop>0:
            decor_char_position = pdb.gimp_image_get_item_position(image, decor_char) # find position of layer in the item tree
            decor_char_copy2 = pdb.gimp_layer_copy(decor_char, True) # copy and insert layer
            pdb.gimp_image_insert_layer(image, decor_char_copy2, None, decor_char_position)
            fill_layer2 = gimp.Layer(image, "Fill Layer2", decor_char.width, decor_char.height, RGBA_IMAGE, 100, darken_mode) # add a fill layer with selected darker mode
            pdb.gimp_drawable_fill(fill_layer2, 0) # fill layer with current shade of grey
            image.add_layer(fill_layer2, -1)
            pdb.gimp_item_set_visible(decor_char_copy2, True) # Make sure decor layer copy is visible
            fill_layer2 = pdb.gimp_image_merge_down(image, fill_layer2, 2) #  and merge down the fill layer onto the decor layer copy
            pdb.gimp_item_set_name(fill_layer2, "Fill Layer2")
            pdb.gimp_layer_set_mode(fill_layer2, 28) # set mode to normal
            pdb.gimp_layer_translate(fill_layer2, ((0+shade_loop)* x_adj)*scale_factor, ((0+shade_loop)*y_adj)*(scale_factor/2))
            # move the layer by 3D amount
            pdb.gimp_image_lower_item_to_bottom(image, fill_layer2)
            if blur>0: pdb.plug_in_gauss(image, fill_layer2, blur, blur, 0)
        else:
            # first loop specifics
            decor_char_position = pdb.gimp_image_get_item_position(image, decor_char) # find position of layer in the item tree
            decor_char_copy = pdb.gimp_layer_copy(decor_char, True) # copy and insert layer
            pdb.gimp_image_insert_layer(image, decor_char_copy, None, decor_char_position)
            fill_layer = gimp.Layer(image, "Fill Layer", decor_char_copy.width, decor_char_copy.height, RGBA_IMAGE, 100, darken_mode) # add a fill layer with selected darker mode
            pdb.gimp_drawable_fill(fill_layer, 0) # fill layer with current shade of grey
            image.add_layer(fill_layer, -1)
            pdb.gimp_item_set_visible(decor_char_copy, True) # Make sure decor layer copy is visible
            fill_layer = pdb.gimp_image_merge_down(image, fill_layer, 2) #  and merge down the fill layer onto the decor layer copy
            pdb.gimp_item_set_name(fill_layer, "Fill Layer")
            pdb.gimp_layer_set_mode(fill_layer, 28) # set mode to normal
            pdb.gimp_layer_translate(fill_layer, x_adj, y_adj) # move the layer
            pdb.gimp_image_lower_item_to_bottom(image, fill_layer)
            if blur>0: pdb.plug_in_gauss(image, fill_layer, blur, blur, 0)
        my_grey -= shade_step # adjust grey shade value
					
        shade_loop += 1
		
    return decor_char

######################################################################################################################################
	
def proc_do_3D_scale(image, decor_char, layer_copy, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_inc, blur, shade_loop, my_grey):

    scale_factor = scale_inc
    translate_inc = scale_factor
    while shade_loop<shade_loop_max:
        new_grey = (1.0/255) * my_grey
        grey_shade = gimpcolor.RGB(new_grey, new_grey, new_grey)
        pdb.gimp_context_set_foreground(grey_shade)
        if shade_loop>0:
                            
            decor_char_position = pdb.gimp_image_get_item_position(image, decor_char) # find position of layer in the item tree
            decor_char_copy2 = pdb.gimp_layer_copy(decor_char, True) # copy and insert layer
            pdb.gimp_image_insert_layer(image, decor_char_copy2, None, decor_char_position)
            fill_layer2 = gimp.Layer(image, "Fill Layer2", decor_char.width, decor_char.height, RGBA_IMAGE, 100, darken_mode) # add a fill layer with selected darker mode
            pdb.gimp_drawable_fill(fill_layer2, 0) # fill layer with current shade of grey
            image.add_layer(fill_layer2, -1)
            pdb.gimp_item_set_visible(decor_char_copy2, True) # Make sure decor layer copy is visible
            fill_layer2 = pdb.gimp_image_merge_down(image, fill_layer2, 2) #  and merge down the fill layer onto the decor layer copy
            pdb.gimp_item_set_name(fill_layer2, "Fill Layer2")
            pdb.gimp_layer_set_mode(fill_layer2, 28) # set mode to normal
            pdb.gimp_layer_scale(fill_layer2,layer_copy.width-scale_factor, layer_copy.height-scale_factor, False)
							
            pdb.gimp_item_transform_translate(fill_layer2, x_adj*scale_factor, y_adj * (scale_factor/2)) # move the layer
            pdb.gimp_image_lower_item_to_bottom(image, fill_layer2)
            if blur>0: pdb.plug_in_gauss(image, fill_layer2, blur, blur, 0)
            err_list.append("scale_factor = " + str(scale_factor) + " : x_adj =" + str((1+shade_loop)* x_adj) )
        else:
            # first loop specifics
            pdb.gimp_item_set_visible(layer_copy, False)
            decor_char_position = pdb.gimp_image_get_item_position(image, decor_char) # find position of layer in the item tree
            decor_char_copy = pdb.gimp_layer_copy(decor_char, True) # copy and insert layer
            pdb.gimp_image_insert_layer(image, decor_char_copy, None, decor_char_position)
            fill_layer = gimp.Layer(image, "Fill Layer", decor_char_copy.width, decor_char_copy.height, RGBA_IMAGE, 100, darken_mode) # add a fill layer with selected darker mode
            pdb.gimp_drawable_fill(fill_layer, 0) # fill layer with current shade of grey
            image.add_layer(fill_layer, -1)
            pdb.gimp_item_set_visible(decor_char_copy, True) # Make sure decor layer copy is visible
            fill_layer = pdb.gimp_image_merge_down(image, fill_layer, 2) #  and merge down the fill layer onto the decor layer copy
            pdb.gimp_item_set_name(fill_layer, "Fill Layer")
            pdb.gimp_layer_set_mode(fill_layer, 28) # set mode to normal
							
            pdb.gimp_layer_scale(fill_layer,fill_layer.width-scale_factor, fill_layer.height-scale_factor, False)
            pdb.gimp_item_transform_translate(fill_layer, x_adj*scale_factor, y_adj * (scale_factor/2)) # move the layer
            pdb.gimp_image_lower_item_to_bottom(image, fill_layer)
            if blur>0: pdb.plug_in_gauss(image, fill_layer, blur, blur, 0)
            err_list = ["scale_factor = " + str(scale_factor) + " : x_adj =" + str(x_adj)]
        my_grey -= shade_step # adjust grey shade value
        scale_factor += scale_inc
        translate_inc += scale_inc
					
        shade_loop += 1
		
		
    return decor_char, layer_copy
	
    #### Useful Functions
# -------------------------------------------------------------------------------------
# -------------------------------------------------------------------------------------

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value


######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################
		
def proc_trim_away(img, layer, amount):
    #### Trim off pixels from a layer
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
    #### End of Proc_trim_away
	

######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Faux_3D_v0-03",
    "Adds 3D effect to custom font characters",
    "Adds 3D effect to custom font characters",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2022",
    "<Image>/Filters/Custom Font Tools/Custom Font Faux 3D v0.03...",
    "RGB*",
    [
    (PF_OPTION, "style_option", "3D style", 0, ("Style 1 - Translated", "Style 2 - Scaled")),
    (PF_SPINNER, "scale_inc_value", "Style scale factor: ", 2, (2, 20, 1)),
    (PF_SPINNER, "depth", "3D depth: ", 8, (2, 400, 1)),
    (PF_SPINNER, "blur", "3D layer blur: ", 1, (1, 20, 1)),
    (PF_TOGGLE, "pf_Use_decor", "Use image as decor face:", 0),
    (PF_SPINNER, "initial_shade", "Initial shade for depth: ", 250, (64,250,1)),
    (PF_SPINNER, "final_shade", "Range for shade: ", 96, (64,96,1)),
    (PF_OPTION, "mode_option", "Darken mode used: ", 0, ("Darken only", "Luma darken", "Multiply", "Burn", "Linear burn")),
    (PF_OPTION, "option", "3D direction", 3, ("North", "North-east", "East", "South-east","South","South-west","West","North-west")),
    (PF_TOGGLE, "adjust_kerning", "Adjust any kerning paths: ", 0),
    (PF_TOGGLE, "do_preview", "Preview single character\n(to a new image): ", 0),
    (PF_STRING, "preview_string", "Preview character: ", "A"),
    ],
    [],
    do_stuff)

main()
